﻿/// Copyright (c) 2021 Iiro Iivanainen, Harri Linna, Jere Pakkanen, Riikka Vilavaara
/// 
/// Permission is hereby granted, free of charge, to any person obtaining
/// a copy of this software and associated documentation files (the
/// "Software"), to deal in the Software without restriction, including
/// without limitation the rights to use, copy, modify, merge, publish,
/// distribute, sublicense, and/or sell copies of the Software, and to
/// permit persons to whom the Software is furnished to do so, subject to
/// the following conditions:
/// 
/// The above copyright notice and this permission notice shall be included
/// in all copies or substantial portions of the Software.
/// 
/// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
/// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
/// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
/// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
/// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
/// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
/// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
/// 

using System.Text.Json.Serialization;

namespace CoreLibrary
{
    /// <summary>
    /// Map class contains one map slice from each axel from specified depth
    /// </summary>
    public class Map
    {
        public Slice StackX { get; set; }
        public Slice StackY { get; set; }
        public Slice StackZ { get; set; }
        public int SlicesX { get; set; }
        public int SlicesY { get; set; }
        public int SlicesZ { get; set; }

        /// <summary>
        /// File type of Map
        /// </summary>
        public enum FileType
        {
            ImageSequence,
            Tiff3D,
            Raw
        }

        /// <summary>
        /// Axis of slice.
        /// </summary>
        public enum Axis
        {
            X,
            Y,
            Z
        }

        public FileType MapType { get; set; }

        public bool LittleEndian { get; set; }
        public int ByteDepth { get; set; }

        /// <summary>
        /// Constructor for JSON serialization and deserialization. Note that the 
        /// parameter names must match the attribute names (non case sensitive)
        /// </summary>
        /// <param name="stackx">Stack in X-axel perspective</param>
        /// <param name="stacky">Stack in Y-axel perspective</param>
        /// <param name="stackz">Stack in Z-axel perspective</param>
        /// <param name="slicesx">Amount of slices in X stack</param>
        /// <param name="slicesy">Amount of slices in Y stack</param>
        /// <param name="slicesz">Amount of slices in Z stack</param>
        /// <param name="maptype">Type of the map</param>
        /// <param name="littleendian">Is the raw file in little endian</param>
        /// <param name="bytedepth">How many bytes reperesent 1 pixel in raw files</param>
        [JsonConstructor]
        public Map(Slice stackx, Slice stacky, Slice stackz, int slicesx, int slicesy, int slicesz, FileType maptype, bool littleendian, int bytedepth)
        {
            StackX = stackx;
            StackY = stacky;
            StackZ = stackz;
            SlicesX = slicesx;
            SlicesY = slicesy;
            SlicesZ = slicesz;
            MapType = maptype;
            LittleEndian = littleendian;
            ByteDepth = bytedepth;
        }


        /// <summary>
        /// Constructor for Map.
        /// </summary>
        public Map()
        {
            StackX = new();
            StackY = new();
            StackZ = new();
        }


        /// <summary>
        /// Returns width of Z stack.
        /// </summary>
        /// <returns>Width of Z stack images.</returns>
        public int GetZWidth()
        {
            return StackZ.Stack.Width;
        }


        /// <summary>
        /// Returns Z stack's height.
        /// </summary>
        /// <returns>Height of Z stack images.</returns>
        public int GetZHeigth()
        {
            return StackZ.Stack.Height;
        }


        /// <summary>
        /// Returns how many slices Z stack has.
        /// </summary>
        /// <returns>Amount of slices in stack Z.</returns>
        public int GetZDepth()
        {
            return SlicesZ;
        }


        /// <summary>
        /// Updates stack Z.
        /// </summary>
        /// <param name="depth">New depth of stack.</param>
        internal void UpdateStackZ(int depth)
        {
            UpdateStack(depth, SlicesX, SlicesY, StackZ);
        }

        
        /// <summary>
        /// Updates stack X.
        /// </summary>
        /// <param name="depth">New depth of stack.</param>
        internal void UpdateStackX(int depth)
        {
            UpdateStack(depth, SlicesZ, SlicesY, StackX);
        }


        /// <summary>
        /// Updates stack Y.
        /// </summary>
        /// <param name="depth">New depth of stack.</param>
        internal void UpdateStackY(int depth)
        {
            UpdateStack(depth, SlicesX, SlicesZ, StackY);
        }


        /// <summary>
        /// Updates specified slice accordingly.
        /// </summary>
        /// <param name="depth">New depth of slice.</param>
        /// <param name="width">Width of stack.</param>
        /// <param name="height">Height of stack.</param>
        /// <param name="slice">Slice to be updated.</param>
        internal void UpdateStack(int depth, int width, int height, Slice slice)
        {
            switch (MapType)
            {
                case FileType.ImageSequence:
                    slice.UpdateImgSeq(depth);
                    break;
                case FileType.Tiff3D:
                    slice.Update3DTiff(depth);
                    break;
                case FileType.Raw:
                    slice.UpdateRaw(depth, width, height, ByteDepth, LittleEndian);
                    break;
                default:
                    break;
            }
        }


        /// <summary>
        /// Returns bit depth of Map.
        /// </summary>
        /// <returns>Map's bit depth.</returns>
        internal int GetBitDepth()
        {
            if (StackZ.Stack != null) return StackZ.Stack.Depth;
            else return 0;
        }


        /// <summary>
        /// Returns Map's file type extension from filepath.
        /// </summary>
        /// <returns>Map's file type.</returns>
        internal string GetMapFileType()
        {
            string path = StackZ.Filepath.GetPath();
            return path[path.LastIndexOf(".")..];
        }

        /// <summary>
        /// Sets filepaths to be in relation of given file
        /// </summary>
        /// <param name="relativeTo">Path that the the filepaths are relative to</param>
        internal void SetMapRelation(string relativeTo)
        {
            if (StackX.Filepath != null) StackX.SetSliceRelation(relativeTo);
            if (StackY.Filepath != null) StackY.SetSliceRelation(relativeTo);
            if (StackZ.Filepath != null) StackZ.SetSliceRelation(relativeTo);
        }
    }
}
