﻿/// Copyright (c) 2021 Iiro Iivanainen, Harri Linna, Jere Pakkanen, Riikka Vilavaara
/// 
/// Permission is hereby granted, free of charge, to any person obtaining
/// a copy of this software and associated documentation files (the
/// "Software"), to deal in the Software without restriction, including
/// without limitation the rights to use, copy, modify, merge, publish,
/// distribute, sublicense, and/or sell copies of the Software, and to
/// permit persons to whom the Software is furnished to do so, subject to
/// the following conditions:
/// 
/// The above copyright notice and this permission notice shall be included
/// in all copies or substantial portions of the Software.
/// 
/// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
/// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
/// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
/// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
/// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
/// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
/// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
/// 

using System.Collections.Generic;
using System.Text.Json.Serialization;

namespace CoreLibrary
{
    /// <summary>
    /// Metas class contains list of metadata.
    /// Contains all metadata that is relative to a single object (attached data or sample)
    /// </summary>
    public class Metas
    {
        public List<Meta> MetaList { get; set; }

        /// <summary>
        /// Constructor for JSON serialization and deserialization. Note that the 
        /// parameter names must match the attribute names (non case sensitive)
        /// </summary>
        /// <param metalist="name"></param>
        [JsonConstructor]
        public Metas(List<Meta> metalist)
        {
            MetaList = metalist;
        }

        /// <summary>
        /// Default constructor of Metas.
        /// </summary>
        public Metas()
        {
            MetaList = new List<Meta>();
        }



        /// <summary>
        /// Returns a clone object of the Metas given
        /// </summary>
        /// <param name="toBeCloned">Metas object to be cloned.</param>
        /// <returns>A clonse object of the metas-object</returns>
        public static Metas Clone(Metas toBeCloned)
        {
            Metas clone = new Metas();

            foreach(Meta m in toBeCloned.MetaList)
            {
                clone.AddData(m.Name, m.Info);
            }

            return clone;
        }


        /// <summary>
        /// Sets metadata of set name to set value or creates metadata
        /// of set name with set value if it doesn't exist.
        /// </summary>
        /// <param name="name">Name of the metadata.</param>
        /// <param name="info">Content of the metadata.</param>
        /// <returns>if the metadata existed.</returns>
        public bool SetData (string name, string info)
        {
            foreach (Meta m in this.MetaList)
            {
                if(m.Name == name)
                {
                    m.Info = info;
                    return true;
                }
            }
            this.AddData(name, info);
            return false;
        }

        /// <summary>
        /// Adds new metadata of specified name and value.
        /// </summary>
        /// <param name="name">Name of the metadata.</param>
        /// <param name="info">Content of the metadata.</param>
        public void AddData(string name, string info)
        {
            var m = new Meta();
            m.Name = name;
            m.Info = info;
            MetaList.Add(m);
        }


        /// <summary>
        /// Adds a metadata object to the metadata list.
        /// </summary>
        /// <param name="meta">Metadata object.</param>
        public void AddData(Meta meta)
        {
            MetaList.Add(meta);
        }


        /// <summary>
        /// Removes the first matching metadata of specified name
        /// </summary>
        /// <param name="name">Name of the metadata.</param>
        /// <returns>Was the metadata found and deleted.</returns>
        public bool RemoveMetaByName(string name)
        {
            var toBeRemoved = MetaList.Find(m => m.Name == name);
            return MetaList.Remove(toBeRemoved);
        }


        /// <summary>
        /// Removes metadata object from metadata list
        /// </summary>
        /// <param name="meta">Metadata object to be removed</param>
        /// <returns>Was the metadata found and deleted.</returns>
        public bool RemoveMeta(Meta meta)
        {
            return MetaList.Remove(meta);
        }


        /// <summary>
        /// Returns the metadata list.
        /// </summary>
        /// <returns>Metadata list.</returns>
        internal List<Meta> GetMetaList()
        {
            return MetaList;
        }
    }
}
