﻿/// Copyright (c) 2021 Iiro Iivanainen, Harri Linna, Jere Pakkanen, Riikka Vilavaara
/// 
/// Permission is hereby granted, free of charge, to any person obtaining
/// a copy of this software and associated documentation files (the
/// "Software"), to deal in the Software without restriction, including
/// without limitation the rights to use, copy, modify, merge, publish,
/// distribute, sublicense, and/or sell copies of the Software, and to
/// permit persons to whom the Software is furnished to do so, subject to
/// the following conditions:
/// 
/// The above copyright notice and this permission notice shall be included
/// in all copies or substantial portions of the Software.
/// 
/// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
/// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
/// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
/// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
/// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
/// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
/// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
/// 

using ImageMagick;
using System.Text.Json.Serialization;

namespace CoreLibrary
{
    /// <summary>
    /// Slice class contains one slice in image form and also the filepath. 
    /// Responsible of updating the image when given a new depth and turning the image to 
    /// byte array.
    /// </summary>
    public class Slice
    {
        [JsonIgnore]
        public MagickImage Stack;

        public RelativeFilepath Filepath { get; set; }

        public string Template { get; set; }

        internal double LevelMin = 0.0;
        internal double LevelMax = 100.0;


        /// <summary>
        /// Constructor for JSON serialization and deserialization. Note that the 
        /// parameter names must match the attribute names (non case sensitive)
        /// </summary>
        /// <param name="filepath"></param>
        /// <param name="template"></param>
        [JsonConstructor]
        public Slice(RelativeFilepath filepath, string template)
        {
            Filepath = filepath;
            Template = template;
        }


        /// <summary>
        /// Default constructor of Slice.
        /// </summary>
        public Slice()
        {
            // Empty.
        }


        /// <summary>
        /// Updates the slice if Map is set as image sequence.
        /// </summary>
        /// <param name="depth">Requested slice</param>
        public void UpdateImgSeq(int depth)
        {
            string temp = "";
            if (Template != null) temp = Template;
            string path = "";
            if (Filepath != null) path = Filepath.GetPath();
            MagickImage img = MapReader.ReadImgSeq(path, depth, temp);
            Stack = img;
        }


        /// <summary>
        /// Updates the slice if Map is set as multiframe tiff.
        /// </summary>
        /// <param name="depth">Requested slice</param>
        internal void Update3DTiff(int depth)
        {
            MagickImage img = MapReader.Read3Dtif(Filepath.GetPath(), depth);
            Stack = img;
        }


        /// <summary>
        /// Updates the slice if Map is set as Raw file type.
        /// </summary>
        /// <param name="depth">Requested slice</param>
        /// <param name="width">Image width</param>
        /// <param name="height">Image height</param>
        /// <param name="byteDepth">Byte depth</param>
        /// <param name="littleEndian">Little endian or not. True if little endian,
        /// false if big endian.</param>
        internal void UpdateRaw(int depth, int width, int height, int byteDepth,
            bool littleEndian)
        {
            MagickImage img = MapReader.ReadRaw(Filepath.GetPath(),
                depth, width, height, byteDepth, littleEndian);
            Stack = img;
        }


        /// <summary>
        /// Return the slice as a byte array. Contains image file.
        /// </summary>
        /// <returns>Byte array of the slice image file.</returns>
        public byte[] GetAsByteArray()
        {
            if (Stack != null)
            {
                var img = new MagickImage(Stack.ToByteArray());
                img.Level(new Percentage(LevelMin), new Percentage(LevelMax));
                return img.ToByteArray();
            }
            return null;
        }


        /// <summary>
        /// Setting file path for the slice.
        /// </summary>
        /// <param name="path">File path.</param>
        public void SetFilepath(string path)
        {
            Filepath = new RelativeFilepath(path);
        }


        /// <summary>
        /// Get file path for the slice.
        /// </summary>
        /// <returns>File path.</returns>
        public string GetFilepath()
        {
            if (Filepath == null) return null;
            return Filepath.GetPath();
        }


        internal void SetSliceRelation(string relativeTo)
        {
            Filepath.SetRelation(relativeTo);
        }


        /// <summary>
        /// Adjust Level of stack.
        /// </summary>
        /// <param name="min">Black point in percentage</param>
        /// <param name="max">White point in percentage</param>
        internal void AdjustLevel(double min, double max)
        {
            LevelMin = min;
            LevelMax = max;
        }
    }
}
